function LambdaError(message) {
	this.message = Error(message).toString();
}

exports.Error = LambdaError;

LambdaError.prototype = Object.create(Error.prototype);

LambdaError.prototype.toString = function () {
	return this.message;
};


function LambdaInternalError() {
	LambdaError.call(this, 'internal error');
}

exports.InternalError = LambdaInternalError;

LambdaInternalError.prototype = Object.create(LambdaError.prototype);


function LambdaSyntaxError() {
	LambdaError.call(this, 'syntax error');
}

exports.SyntaxError = LambdaSyntaxError;

LambdaSyntaxError.prototype = Object.create(LambdaError.prototype);


function LambdaRuntimeError() {
	LambdaError.call(this, 'runtime error');
}

exports.RuntimeError = LambdaRuntimeError;

LambdaRuntimeError.prototype = Object.create(LambdaError.prototype);


function LambdaUserError(value) {
	LambdaError.call(this, 'user error: ' + value.toString());
	this.value = value;
}

exports.UserError = LambdaUserError;

LambdaUserError.prototype = Object.create(LambdaError.prototype);

Array.prototype.contains = function () {
	return this.indexOf.apply(this, arguments) >= 0;
};

Array.prototype.unique = function () {
	var array = [];
	for (var i = 0; i < this.length; i++) {
		if (array.indexOf(this[i]) < 0) {
			array.push(this[i]);
		}
	}
	return array;
};

Array.prototype.union = function () {
	return this.concat.apply(this, arguments).unique();
};

function getGlobalValue(name, fallback) {
	try {
		/*jshint evil: true */
		return eval(name);
	} catch (e) {
		if (fallback) {
			return fallback();
		} else {
			throw e;
		}
	}
}

function Context(hash) {
	this.hash = Object.create(null);
	if (hash) {
		for (var name in hash) {
			if (hash.hasOwnProperty(name)) {
				this.hash[name] = hash[name];
			}
		}
	}
}

exports.Context = Context;

Context.prototype.has = function (name) {
	return name in this.hash;
};

Context.prototype.top = function (name) {
	if (name in this.hash) {
		return this.hash[name];
	} else {
		throw new LambdaInternalError();
	}
};

Context.prototype.forEach = function (callback, context) {
	for (var name in this.hash) {
		callback.call(context, name, this.hash[name]);
	}
};

Context.prototype.add = function (name, value) {
	function Hash() {}
	Hash.prototype = this.hash;
	var hash = new Hash();
	hash[name] = value;
	var context = new Context();
	context.hash = hash;
	return context;
};

Context.prototype.addAll = function (hash) {
	function Hash() {}
	Hash.prototype = this.hash;
	var child = new Hash();
	for (var name in hash) {
		if (hash.hasOwnProperty(name)) {
			child[name] = hash[name];
		}
	}
	var context = new Context();
	context.hash = child;
	return context;
};

Context.EMPTY = new Context();

function AbstractValue() {}

exports.AbstractValue = AbstractValue;

AbstractValue.prototype.is = function (Class) {
	return this instanceof Class;
};

AbstractValue.prototype.isAny = function () {
	for (var i = 0; i < arguments.length; i++) {
		if (this instanceof arguments[i]) {
			return true;
		}
	}
	return false;
};

AbstractValue.prototype.bindThis = function () {
	return this;
};


function UndefinedValue() {
	AbstractValue.call(this);
}

exports.UndefinedValue = UndefinedValue;

UndefinedValue.prototype = Object.create(AbstractValue.prototype);

UndefinedValue.prototype.type = 'undefined';

UndefinedValue.prototype.toString = function () {
	return 'undefined';
};

UndefinedValue.prototype.marshal = function () {};

UndefinedValue.INSTANCE = new UndefinedValue();


function NullValue() {
	AbstractValue.call(this);
}

exports.NullValue = NullValue;

NullValue.prototype = Object.create(AbstractValue.prototype);

NullValue.prototype.type = 'null';

NullValue.prototype.toString = function () {
	return 'null';
};

NullValue.prototype.marshal = function () {
	return null;
};

NullValue.INSTANCE = new NullValue();


function BooleanValue(value) {
	AbstractValue.call(this);
	this.value = !!value;
}

exports.BooleanValue = BooleanValue;

BooleanValue.prototype = Object.create(AbstractValue.prototype);

BooleanValue.prototype.type = 'bool';

BooleanValue.prototype.toString = function () {
	if (this.value) {
		return 'true';
	} else {
		return 'false';
	}
};

BooleanValue.prototype.marshal = function () {
	return this.value;
};

BooleanValue.TRUE = new BooleanValue(true);
BooleanValue.FALSE = new BooleanValue(false);

BooleanValue.unmarshal = function (value) {
	if (value) {
		return BooleanValue.TRUE;
	} else {
		return BooleanValue.FALSE;
	}
};


function IntegerValue(value) {
	AbstractValue.call(this);
	this.value = ~~value;
}

exports.IntegerValue = IntegerValue;

IntegerValue.prototype = Object.create(AbstractValue.prototype);

IntegerValue.prototype.type = 'int';

IntegerValue.prototype.toString = function () {
	return '' + this.value;
};

IntegerValue.prototype.marshal = function () {
	return this.value;
};


function FloatValue(value) {
	AbstractValue.call(this);
	this.value = value * 1;
}

exports.FloatValue = FloatValue;

FloatValue.prototype = Object.create(AbstractValue.prototype);

FloatValue.prototype.type = 'float';

FloatValue.prototype.toString = function () {
	return '' + this.value;
};

FloatValue.prototype.marshal = function () {
	return this.value;
};


function NativeComplexValue(real, imaginary) {
	this.r = real;
	this.i = imaginary;
}

NativeComplexValue.prototype.toString = function () {
	if (this.i < 0) {
		return this.r + '-' + -this.i + 'i';
	} else {
		return this.r + '+' + this.i + 'i';
	}
};


function ComplexValue(real, imaginary) {
	AbstractValue.call(this);
	real = real * 1;
	imaginary = imaginary * 1;
	this.real = real;
	this.imaginary = imaginary;
	this.prototype = new Context({
		real: new FloatValue(real),
		imaginary: new FloatValue(imaginary)
	});
}

exports.ComplexValue = ComplexValue;

ComplexValue.prototype = Object.create(AbstractValue.prototype);

ComplexValue.prototype.type = 'complex';

ComplexValue.prototype.toString = function () {
	if (this.imaginary < 0) {
		return this.real + '-' + -this.imaginary + 'i';
	} else {
		return this.real + '+' + this.imaginary + 'i';
	}
};

ComplexValue.prototype.marshal = function () {
	return new NativeComplexValue(this.real, this.imaginary);
};


function Closure(lambda, context) {
	AbstractValue.call(this);
	this.lambda = lambda;
	this.context = context;
}

exports.Closure = Closure;

Closure.prototype = Object.create(AbstractValue.prototype);

Closure.prototype.type = 'closure';

Closure.prototype.toString = function () {
	return 'closure';
};

Closure.prototype.bindThis = function (value) {
	return new Closure(this.lambda, this.context.add('this', value));
};

Closure.prototype.marshal = function () {
	var length = 0;
	for (var node = this.lambda; node.is(LambdaNode); node = node.body) {
		length++;
	}
	node = this.lambda;
	var context = this.context;
	return function () {
		var values = arguments;
		return (function augment(node, context, index) {
			if (index < length) {
				return augment(node.body, context.add(node.name, AbstractValue.unmarshal(values[index])), index + 1);
			} else {
				return (function () {
					try {
						return node.evaluate(context);
					} catch (e) {
						if (e instanceof LambdaUserError) {
							throw e.value.marshal();
						} else {
							throw e;
						}
					}
				}()).marshal();
			}
		}(node, context, 0));
	};
};

Closure.unmarshal = function (value, context) {
	return new Closure((function makeLambda(index, names) {
		if (index < Math.max(value.length, 1)) {
			var name = '' + index;
			names.push(name);
			return new LambdaNode(name, makeLambda(index + 1, names));
		} else {
			return new NativeNode(value, names);
		}
	}(0, [])), context || Context.EMPTY);
};

Closure.prototype.getLength = function () {
	var length = 0;
	for (var node = this.lambda; node.is(LambdaNode); node = node.body) {
		length++;
	}
	return length;
};


function LazyValue(expression, context) {
	AbstractValue.call(this);
	this.expression = expression;
	this.context = context;
}

exports.LazyValue = LazyValue;

LazyValue.prototype = Object.create(AbstractValue.prototype);

LazyValue.prototype.type = 'lazy';

LazyValue.prototype.toString = function () {
	return 'lazy';
};

LazyValue.prototype.bindThis = function (value) {
	return new LazyValue(this.expression, this.context.add('this', value));
};

LazyValue.prototype.marshal = function () {
	var node = this.expression;
	var context = this.context;
	return function () {
		return (function () {
			try {
				return node.evaluate(context);
			} catch (e) {
				if (e instanceof LambdaUserError) {
					throw e.value.marshal();
				} else {
					throw e;
				}
			}
		}()).marshal();
	};
};

LazyValue.unmarshal = function (value, context) {
	return new LazyValue(new NativeNode(value, []), context || Context.EMPTY);
};

LazyValue.evaluate = function (value) {
	if (value.is(LazyValue)) {
		return value.expression.evaluate(value.context);
	} else {
		return value;
	}
};


function StringValue(value) {
	AbstractValue.call(this);
	this.value = '' + value;
	this.prototype = new Context({
		length: new IntegerValue(value.length)
	});
}

exports.StringValue = StringValue;

StringValue.prototype = Object.create(AbstractValue.prototype);

StringValue.prototype.type = 'string';

StringValue.prototype.toString = function () {
	return '\"' + this.value.replace(/\\/g, '\\\\').replace(/\"/g, '\\\"') + '\"';
};

StringValue.prototype.marshal = function () {
	return this.value;
};


function ArrayValue(array) {
	AbstractValue.call(this);
	this.array = array = array || [];
}

exports.ArrayValue = ArrayValue;

ArrayValue.prototype = Object.create(AbstractValue.prototype);

ArrayValue.prototype.type = 'array';

ArrayValue.prototype.toString = function () {
	return '{ ' + this.array.map(function (element) {
		return element.toString();
	}).join(', ') + ' }';
};

ArrayValue.prototype.marshal = function () {
	return this.array.map(function (value) {
		return value.marshal();
	});
};


function NativeArrayValue(array) {
	AbstractValue.call(this);
	this.array = array;
}

exports.NativeArrayValue = NativeArrayValue;

NativeArrayValue.prototype = Object.create(AbstractValue.prototype);

NativeArrayValue.prototype.type = 'array';

NativeArrayValue.prototype.toString = function () {
	return '{ ' + this.array.map(function (element) {
		return AbstractValue.unmarshal(element).toString();
	}).join(', ') + ' }';
};

NativeArrayValue.prototype.marshal = function () {
	return this.array;
};


function ObjectValue(context) {
	AbstractValue.call(this);
	this.context = context || Context.EMPTY;
}

exports.ObjectValue = ObjectValue;

ObjectValue.prototype = Object.create(AbstractValue.prototype);

ObjectValue.prototype.type = 'object';

ObjectValue.prototype.toString = function () {
	return 'object';
};

ObjectValue.prototype.marshal = function () {
	var object = {};
	this.context.forEach(function (name, value) {
		object[name] = value.marshal();
	});
	return object;
};


function NativeContext(object) {
	this.object = object;
}

exports.NativeContext = NativeContext;

NativeContext.prototype.has = function (name) {
	return name in this.object;
};

NativeContext.prototype.top = function (name) {
	return AbstractValue.unmarshal(this.object[name]);
};

NativeContext.prototype.add = function (name, value) {
	function NativeObject() {}
	NativeObject.prototype = this.object;
	var object = new NativeObject();
	object[name] = value.marshal();
	return new NativeContext(object);
};


function NativeObjectValue(context) {
	AbstractValue.call(this);
	this.context = context;
}

exports.NativeObjectValue = NativeObjectValue;

NativeObjectValue.prototype = Object.create(AbstractValue.prototype);

NativeObjectValue.prototype.type = 'object';

NativeObjectValue.prototype.toString = function () {
	return 'object';
};

NativeObjectValue.prototype.marshal = function () {
	return this.context.object;
};


AbstractValue.unmarshal = function (value) {
	switch (typeof value) {
	case 'undefined':
		return UndefinedValue.INSTANCE;
	case 'boolean':
		return BooleanValue.unmarshal(value);
	case 'number':
		if (value % 1) {
			return new FloatValue(value);
		} else {
			return new IntegerValue(~~value);
		}
		break;
	case 'string':
		return new StringValue(value);
	case 'function':
		return Closure.unmarshal(value);
	case 'object':
		if (value === null) {
			return NullValue.INSTANCE;
		} else if (Array.isArray(value)) {
			return new NativeArrayValue(value);
		} else if (value instanceof NativeComplexValue) {
			return new ComplexValue(value.r, value.i);
		} else {
			return new NativeObjectValue(new NativeContext(value));
		}
		break;
	}
};

function AbstractNode() {}

exports.AbstractNode = AbstractNode;

AbstractNode.prototype.is = function (Class) {
	return this instanceof Class;
};

AbstractNode.prototype.isAny = function () {
	for (var i = 0; i < arguments.length; i++) {
		if (this instanceof arguments[i]) {
			return true;
		}
	}
	return false;
};

AbstractNode.prototype.compile = function () {
	return '(function(){"use strict";' + this.compileStatement() + '}());';
};


function LiteralNode(value) {
	AbstractNode.call(this);
	this.value = value;
}

exports.LiteralNode = LiteralNode;

LiteralNode.prototype = Object.create(AbstractNode.prototype);

LiteralNode.prototype.getFreeVariables = function () {
	return [];
};

LiteralNode.prototype.evaluate = function () {
	return this.value;
};

LiteralNode.prototype.compileExpression = function () {
	return JSON.stringify(this.value.marshal());
};

LiteralNode.prototype.compileStatement = function () {
	return 'return ' + JSON.stringify(this.value.marshal()) + ';';
};


function ArrayLiteralNode(expressions) {
	AbstractNode.call(this);
	this.expressions = expressions;
}

exports.ArrayLiteralNode = ArrayLiteralNode;

ArrayLiteralNode.prototype = Object.create(AbstractNode.prototype);

ArrayLiteralNode.prototype.getFreeVariables = function () {
	var names = [];
	this.expressions.forEach(function (expression) {
		names = names.union(expression.getFreeVariables());
	});
	return names;
};

ArrayLiteralNode.prototype.evaluate = function (context) {
	return new ArrayValue(this.expressions.map(function (expression) {
		return expression.evaluate(context);
	}));
};

ArrayLiteralNode.prototype.compileExpression = function () {
	return '[' + this.expressions.map(function (expression) {
		return expression.compileExpression();
	}).join(',') + ']';
};

ArrayLiteralNode.prototype.compileStatement = function () {
	return 'return[' + this.expressions.map(function (expression) {
		return expression.compileExpression();
	}).join(',') + '];';
};


function VariableNode(name) {
	AbstractNode.call(this);
	this.name = name;
}

exports.VariableNode = VariableNode;

VariableNode.prototype = Object.create(AbstractNode.prototype);

VariableNode.prototype.getFreeVariables = function () {
	return [this.name];
};

VariableNode.prototype.evaluate = function (context) {
	if (context.has(this.name)) {
		return LazyValue.evaluate(context.top(this.name));
	} else {
		return LazyValue.evaluate(AbstractValue.unmarshal(getGlobalValue(this.name, function () {
			throw new LambdaRuntimeError();
		})));
	}
};

VariableNode.prototype.compileExpression = function () {
	return this.name;
};

VariableNode.prototype.compileStatement = function () {
	return 'return ' + this.name + ';';
};


function FixNode() {
	AbstractNode.call(this);
}

exports.FixNode = FixNode;

FixNode.prototype = Object.create(AbstractNode.prototype);

FixNode.prototype.getFreeVariables = function () {
	return [];
};

FixNode.prototype.evaluate = function () {
	return FixNode.Z_COMBINATOR;
};

FixNode.prototype.compileExpression = function () {
	return 'function fix(f){return function(v){return f(fix(f))(v);};}';
};

FixNode.prototype.compileStatement = function () {
	return 'return function fix(f){return function(v){return f(fix(f))(v);};};';
};


function ThisNode() {
	AbstractNode.call(this);
}

exports.ThisNode = ThisNode;

ThisNode.prototype = Object.create(AbstractNode.prototype);

ThisNode.prototype.getFreeVariables = function () {
	return ['this'];
};

ThisNode.prototype.evaluate = function (context) {
	if (context.has('this')) {
		return context.top('this');
	} else {
		throw new LambdaRuntimeError();
	}
};

ThisNode.prototype.compileExpression = function () {
	return 'this';
};

ThisNode.prototype.compileStatement = function () {
	return 'return this;';
};


function ErrorNode() {
	AbstractNode.call(this);
}

exports.ErrorNode = ErrorNode;

ErrorNode.prototype = Object.create(AbstractNode.prototype);

ErrorNode.prototype.getFreeVariables = function () {
	return ['error'];
};

ErrorNode.prototype.evaluate = function (context) {
	if (context.has('error')) {
		return context.top('error');
	} else {
		throw new LambdaRuntimeError();
	}
};

ErrorNode.prototype.compileExpression = function () {
	return 'error';
};

ErrorNode.prototype.compileStatement = function () {
	return 'return error;';
};


function FieldAccessNode(left, name) {
	AbstractNode.call(this);
	this.left = left;
	this.name = name;
}

exports.FieldAccessNode = FieldAccessNode;

FieldAccessNode.prototype = Object.create(AbstractNode.prototype);

FieldAccessNode.prototype.getFreeVariables = function () {
	return this.left.getFreeVariables();
};

FieldAccessNode.prototype.evaluate = function (context) {
	var left = LazyValue.evaluate(this.left.evaluate(context));
	if (left.isAny(ObjectValue, NativeObjectValue) && left.context.has(this.name)) {
		return LazyValue.evaluate(left.context.top(this.name).bindThis(left));
	} else if (left.isAny(ComplexValue, StringValue, ArrayValue, NativeArrayValue, Closure) && left.prototype.has(this.name)) {
		return LazyValue.evaluate(left.prototype.top(this.name).bindThis(left));
	} else {
		throw new LambdaRuntimeError();
	}
};

FieldAccessNode.prototype.compileExpression = function () {
	return '(' + this.left.compileExpression() + ').' + this.name;
};

FieldAccessNode.prototype.compileStatement = function () {
	return 'return(' + this.left.compileExpression() + ').' + this.name + ';';
};


function SubscriptNode(expression, index) {
	AbstractNode.call(this);
	this.expression = expression;
	this.index = index;
}

exports.SubscriptNode = SubscriptNode;

SubscriptNode.prototype = Object.create(AbstractNode.prototype);

SubscriptNode.prototype.getFreeVariables = function () {
	return this.expression.getFreeVariables().union(this.index.getFreeVariables());
};

SubscriptNode.prototype.evaluate = function (context) {
	var value = LazyValue.evaluate(this.expression.evaluate(context));
	if (value.isAny(ArrayValue, NativeArrayValue, StringValue)) {
		var index = this.index.evaluate(context);
		if (index.is(IntegerValue)) {
			if (value.is(ArrayValue)) {
				if (index.value >= 0 && index.value < value.array.length) {
					return LazyValue.evaluate(value.array[index.value]);
				}
			} else if (value.is(NativeArrayValue)) {
				if (index.value >= 0 && index.value < value.array.length) {
					return LazyValue.evaluate(AbstractValue.unmarshal(value.array[index.value]));
				}
			} else if (value.is(StringValue)) {
				if (index.value >= 0 && index.value < value.value.length) {
					return LazyValue.evaluate(value.value[index.value]);
				}
			}
		}
	}
	throw new LambdaRuntimeError();
};

SubscriptNode.prototype.compileExpression = function () {
	return '(' + this.expression.compileExpression() + ')[' + this.index.compileExpression() + ']';
};

SubscriptNode.prototype.compileStatement = function () {
	return 'return(' + this.expression.compileExpression() + ')[' + this.index.compileExpression() + '];';
};


function LambdaNode(name, body) {
	AbstractNode.call(this);
	this.name = name;
	this.body = body;
}

exports.LambdaNode = LambdaNode;

LambdaNode.prototype = Object.create(AbstractNode.prototype);

LambdaNode.prototype.getFreeVariables = function () {
	return this.body.getFreeVariables().filter(function (name) {
		return name !== this.name;
	}, this);
};

LambdaNode.prototype.evaluate = function (context) {
	return new Closure(this, context);
};

LambdaNode.prototype.compileExpression = function () {
	return 'function(' + this.name + '){' + this.body.compileStatement() + '}';
};

LambdaNode.prototype.compileStatement = function () {
	return 'return function(' + this.name + '){' + this.body.compileStatement() + '};';
};


function LazyNode(expression) {
	AbstractNode.call(this);
	this.expression = expression;
}

exports.LazyNode = LazyNode;

LazyNode.prototype = Object.create(AbstractNode.prototype);

LazyNode.prototype.getFreeVariables = function () {
	return this.expression.getFreeVariables();
};

LazyNode.prototype.evaluate = function (context) {
	return new LazyValue(this.expression, context);
};

LazyNode.prototype.compileExpression = function () {
	return 'function(){' + this.expression.compileStatement() + '}';
};

LazyNode.prototype.compileStatement = function () {
	return 'return function(){' + this.expression.compileStatement() + '}';
};


function ApplicationNode(left, right) {
	AbstractNode.call(this);
	this.left = left;
	this.right = right;
}

exports.ApplicationNode = ApplicationNode;

ApplicationNode.prototype = Object.create(AbstractNode.prototype);

ApplicationNode.prototype.getFreeVariables = function () {
	return this.left.getFreeVariables().union(this.right.getFreeVariables());
};

ApplicationNode.prototype.evaluate = function (context) {
	var left = this.left.evaluate(context);
	if (left.is(Closure)) {
		return left.lambda.body.evaluate(left.context.add(left.lambda.name, this.right.evaluate(context)));
	} else {
		throw new LambdaRuntimeError();
	}
};

ApplicationNode.prototype.compileExpression = function () {
	return '(' + this.left.compileExpression() + ')(' + this.right.compileExpression() + ')';
};

ApplicationNode.prototype.compileStatement = function () {
	return 'return(' + this.left.compileExpression() + ')(' + this.right.compileExpression() + ');';
};


function LetNode(names, expression, body) {
	AbstractNode.call(this);
	this.names = names;
	this.expression = expression;
	this.body = body;
}

exports.LetNode = LetNode;

LetNode.prototype = Object.create(AbstractNode.prototype);

LetNode.prototype.getFreeVariables = function () {
	return this.expression.getFreeVariables().union(this.body.getFreeVariables().filter(function (name) {
		return name !== this.names[0];
	}, this));
};

LetNode.prototype.evaluate = function (context) {
	var names = this.names;
	var value = this.expression.evaluate(context);
	return this.body.evaluate((function augment(context, index) {
		if (index < names.length - 1) {
			var name = names[index];
			if (context.has(name)) {
				var object = context.top(name);
				if (object.is(ObjectValue)) {
					return context.add(name, new ObjectValue(augment(object.context, index + 1)));
				} else if (object.is(NativeObjectValue)) {
					return context.add(name, new NativeObjectValue(augment(object.context, index + 1)));
				} else {
					return context.add(name, new ObjectValue(augment(Context.EMPTY, index + 1)));
				}
			} else {
				return augment(context.add(name, (function () {
					try {
						return AbstractValue.unmarshal(getGlobalValue(name));
					} catch (e) {
						return new ObjectValue(Context.EMPTY);
					}
				}())), index);
			}
		} else if (index < names.length) {
			return context.add(names[index], value);
		} else {
			throw new LambdaInternalError();
		}
	}(context, 0)));
};

LetNode.prototype.compileExpression = function () {
	if (this.names.length > 1) {
		// XXX not implemented yet
		throw new LambdaInternalError();
	} else {
		return '(function(' + this.names[0] + '){' + this.body.compileStatement() + '}(' + this.expression.compileExpression() + '))';
	}
};

LetNode.prototype.compileStatement = function () {
	if (this.names.length > 1) {
		// XXX not implemented yet
		throw new LambdaInternalError();
	} else {
		return 'var ' + this.names[0] + '=' + this.expression.compileExpression() + ';' + this.body.compileStatement();
	}
};


function IfNode(condition, thenExpression, elseExpression) {
	AbstractNode.call(this);
	this.condition = condition;
	this.thenExpression = thenExpression;
	this.elseExpression = elseExpression;
}

exports.IfNode = IfNode;

IfNode.prototype = Object.create(AbstractNode.prototype);

IfNode.prototype.getFreeVariables = function () {
	return this.condition.getFreeVariables()
		.union(this.thenExpression.getFreeVariables())
		.union(this.elseExpression.getFreeVariables());
};

IfNode.prototype.evaluate = function (context) {
	var condition = this.condition.evaluate(context);
	if (condition.is(BooleanValue)) {
		if (condition.value) {
			return this.thenExpression.evaluate(context);
		} else {
			return this.elseExpression.evaluate(context);
		}
	} else {
		throw new LambdaRuntimeError();
	}
};

IfNode.prototype.compileExpression = function () {
	return '(' + this.condition.compileExpression() + ')?(' + this.thenExpression.compileExpression() + '):(' + this.elseExpression.compileExpression() + ')';
};

IfNode.prototype.compileStatement = function () {
	return 'if(' + this.condition.compileExpression() + '){' + this.thenExpression.compileStatement() + '}else{' + this.elseExpression.compileStatement() + '}';
};


function ThrowNode(expression) {
	AbstractNode.call(this);
	this.expression = expression;
}

exports.ThrowNode = ThrowNode;

ThrowNode.prototype = Object.create(AbstractNode.prototype);

ThrowNode.prototype.getFreeVariables = function () {
	return this.expression.getFreeVariables();
};

ThrowNode.prototype.evaluate = function (context) {
	throw new LambdaUserError(this.expression.evaluate(context));
};

ThrowNode.prototype.compileExpression = function () {
	return '(function(){throw ' + this.expression.compileExpression() + ';}())';
};

ThrowNode.prototype.compileStatement = function () {
	return 'throw ' + this.expression.compileExpression() + ';';
};


function TryCatchNode(tryExpression, catchExpression) {
	AbstractNode.call(this);
	this.tryExpression = tryExpression;
	this.catchExpression = catchExpression;
}

exports.TryCatchNode = TryCatchNode;

TryCatchNode.prototype = Object.create(AbstractNode.prototype);

TryCatchNode.prototype.getFreeVariables = function () {
	return this.tryExpression.getFreeVariables()
		.union(this.catchExpression.getFreeVariables().filter(function (name) {
			return name !== 'error';
	}));
};

TryCatchNode.prototype.evaluate = function (context) {
	try {
		return this.tryExpression.evaluate(context);
	} catch (e) {
		if (e instanceof LambdaUserError) {
			return this.catchExpression.evaluate(context.add('error', e.value));
		} else {
			throw e;
		}
	}
};

TryCatchNode.prototype.compileExpression = function () {
	return '(function(){try{' + this.tryExpression.compileStatement() + '}catch(error){' + this.catchExpression.compileStatement() + '}}())';
};

TryCatchNode.prototype.compileStatement = function () {
	return 'try{' + this.tryExpression.compileStatement() + '}catch(error){' + this.catchExpression.compileStatement() + '}';
};


function TryFinallyNode(tryExpression, finallyExpression) {
	AbstractNode.call(this);
	this.tryExpression = tryExpression;
	this.finallyExpression = finallyExpression;
}

exports.TryFinallyNode = TryFinallyNode;

TryFinallyNode.prototype = Object.create(AbstractNode.prototype);

TryFinallyNode.prototype.getFreeVariables = function () {
	return this.tryExpression.getFreeVariables()
		.union(this.finallyExpression.getFreeVariables());
};

TryFinallyNode.prototype.evaluate = function (context) {
	try {
		return this.tryExpression.evaluate(context);
	} finally {
		this.finallyExpression.evaluate(context);
	}
};

TryFinallyNode.prototype.compileExpression = function () {
	return '(function(){try{' + this.tryExpression.compileStatement() + '}finally{' + this.finallyExpression.compileStatement() + '}}())';
};

TryFinallyNode.prototype.compileStatement = function () {
	return 'try{' + this.tryExpression.compileStatement() + '}finally{' + this.finallyExpression.compileStatement() + '}';
};


function TryCatchFinallyNode(tryExpression, catchExpression, finallyExpression) {
	AbstractNode.call(this);
	this.tryExpression = tryExpression;
	this.catchExpression = catchExpression;
	this.finallyExpression = finallyExpression;
}

exports.TryCatchFinallyNode = TryCatchFinallyNode;

TryCatchFinallyNode.prototype = Object.create(AbstractNode.prototype);

TryCatchFinallyNode.prototype.getFreeVariables = function () {
	return this.tryExpression.getFreeVariables()
		.union(this.catchExpression.getFreeVariables().filter(function (name) {
			return name !== 'error';
	}))
		.union(this.finallyExpression.getFreeVariables());
};

TryCatchFinallyNode.prototype.evaluate = function (context) {
	try {
		return this.tryExpression.evaluate(context);
	} catch (e) {
		if (e instanceof LambdaUserError) {
			return this.catchExpression.evaluate(context.add('error', e.value));
		} else {
			throw e;
		}
	} finally {
		this.finallyExpression.evaluate(context);
	}
};

TryCatchFinallyNode.prototype.compileExpression = function () {
	return '(function(){try{' + this.tryExpression.compileStatement() +
		'}catch(error){' + this.catchExpression.compileStatement() +
		'}finally{' + this.finallyExpression.compileStatement() + '}}())';
};

TryCatchFinallyNode.prototype.compileStatement = function () {
	return 'try{' + this.tryExpression.compileStatement() +
		'}catch(error){' + this.catchExpression.compileStatement() +
		'}finally{' + this.finallyExpression.compileStatement() + '}';
};


function NativeNode(nativeFunction, argumentNames) {
	AbstractNode.call(this);
	this.nativeFunction = nativeFunction;
	this.argumentNames = argumentNames;
}

exports.NativeNode = NativeNode;

NativeNode.prototype = Object.create(AbstractNode.prototype);

NativeNode.prototype.getFreeVariables = function () {
	return this.argumentNames;
};

NativeNode.prototype.evaluate = function (context) {
	var thisValue = (function () {
		if (context.has('this')) {
			return context.top('this').marshal();
		} else {
			return null;
		}
	}());
	var argumentValues = this.argumentNames.map(function (name) {
		if (context.has(name)) {
			return context.top(name).marshal();
		} else {
			throw new LambdaInternalError();
		}
	});
	return AbstractValue.unmarshal(function () {
		try {
			return this.nativeFunction.apply(thisValue, argumentValues);
		} catch (e) {
			if (e instanceof LambdaError) {
				throw e;
			} else {
				throw new LambdaUserError(AbstractValue.unmarshal(e));
			}
		}
	}.call(this));
};


function SemiNativeNode(overloads, ariety) {
	AbstractNode.call(this);
	this.overloads = overloads;
	this.argumentNames = [];
	for (var i = 0; i < ariety; i++) {
		this.argumentNames.push('' + i);
	}
}

exports.SemiNativeNode = SemiNativeNode;

SemiNativeNode.prototype = Object.create(AbstractNode.prototype);

SemiNativeNode.prototype.getFreeVariables = function () {
	return this.argumentNames;
};

SemiNativeNode.prototype.evaluate = function (context) {
	var argumentValues = this.argumentNames.map(function (name) {
		if (context.has(name)) {
			return context.top(name);
		} else {
			throw new LambdaInternalError();
		}
	});
	var overload = this.overloads;
	argumentValues.forEach(function (argument) {
		for (var key in overload) {
			if (overload.hasOwnProperty(key)) {
				if ((new RegExp('^(' + key + ')$')).test(argument.type)) {
					overload = overload[key];
					return;
				}
			}
		}
		throw new LambdaRuntimeError();
	});
	return overload.apply(null, argumentValues);
};


function UnaryOperatorNode(overloads) {
	LambdaNode.call(this, '0', new SemiNativeNode(overloads, 1));
}

exports.UnaryOperatorNode = UnaryOperatorNode;

UnaryOperatorNode.prototype = Object.create(LambdaNode.prototype);


function BinaryOperatorNode(overloads) {
	LambdaNode.call(this, '0', new LambdaNode('1', new SemiNativeNode(overloads, 2)));
}

exports.BinaryOperatorNode = BinaryOperatorNode;

BinaryOperatorNode.prototype = Object.create(LambdaNode.prototype);

function TypeOfOperator() {
	UnaryOperatorNode.call(this, {
		'.*': function (x) {
			return new StringValue(x.type);
		}
	});
}

exports.TypeOfOperator = TypeOfOperator;

TypeOfOperator.prototype = Object.create(UnaryOperatorNode.prototype);


function LogicalNotOperator() {
	UnaryOperatorNode.call(this, {
		'bool': function (x) {
			return BooleanValue.unmarshal(!x.value);
		}
	});
}

exports.LogicalNotOperator = LogicalNotOperator;

LogicalNotOperator.prototype = Object.create(UnaryOperatorNode.prototype);


function BitwiseNotOperator() {
	UnaryOperatorNode.call(this, {
		'int': function (x) {
			return new IntegerValue(~x.value);
		}
	});
}

exports.BitwiseNotOperator = BitwiseNotOperator;

BitwiseNotOperator.prototype = Object.create(UnaryOperatorNode.prototype);


function PlusOperator() {
	BinaryOperatorNode.call(this, {
		'undefined': {
			'string': function (x, y) {
				return new StringValue('undefined' + y.value);
			}
		},
		'null': {
			'string': function (x, y) {
				return new StringValue('null' + y.value);
			}
		},
		'bool': {
			'bool': function (x, y) {
				return BooleanValue.unmarshal(x.value, y.value);
			},
			'string': function (x, y) {
				if (x.value) {
					return new StringValue('true' + y.value);
				} else {
					return new StringValue('false' + y.value);
				}
			}
		},
		'int': {
			'int': function (x, y) {
				return new IntegerValue(x.value + y.value);
			},
			'float': function (x, y) {
				return new FloatValue(x.value + y.value);
			},
			'complex': function (x, y) {
				return new ComplexValue(x.value + y.real, y.imaginary);
			},
			'string': function (x, y) {
				return new StringValue(('' + x.value) + y.value);
			}
		},
		'float': {
			'int|float': function (x, y) {
				return new FloatValue(x.value + y.value);
			},
			'complex': function (x, y) {
				return new ComplexValue(x.value + y.real, y.imaginary);
			},
			'string': function (x, y) {
				return new StringValue(('' + x.value) + y.value);
			}
		},
		'complex': {
			'int|float': function (x, y) {
				return new ComplexValue(x.real + y.value, x.imaginary);
			},
			'complex': function (x, y) {
				return new ComplexValue(x.real + y.real, x.imaginary + y.imaginary);
			},
			'string': function (x, y) {
				return new StringValue(x.toString() + y.value);
			}
		},
		'string': {
			'undefined': function (x) {
				return new StringValue(x.value + 'undefined');
			},
			'null': function (x) {
				return new StringValue(x.value + 'null');
			},
			'bool': function (x, y) {
				if (y.value) {
					return new StringValue(x.value + 'true');
				} else {
					return new StringValue(x.value + 'false');
				}
			},
			'int|float': function (x, y) {
				return new StringValue(x.value + y.value);
			},
			'complex': function (x, y) {
				return new StringValue(x.value + y.toString());
			},
			'string': function (x, y) {
				return new StringValue(x.value + y.value);
			}
		},
		'array': {
			'array': function (x, y) {
				return new ArrayValue(x.array.concat(y.array));
			}
		}
	});
}

exports.PlusOperator = PlusOperator;

PlusOperator.prototype = Object.create(BinaryOperatorNode.prototype);


function MinusOperator() {
	BinaryOperatorNode.call(this, {
		'bool': {
			'bool': function (x, y) {
				return BooleanValue.unmarshal(x.value && !y.value);
			}
		},
		'int': {
			'int': function (x, y) {
				return new IntegerValue(x.value - y.value);
			},
			'float': function (x, y) {
				return new FloatValue(x.value - y.value);
			},
			'complex': function (x, y) {
				return new ComplexValue(x.value - y.real, -y.imaginary);
			}
		},
		'float': {
			'int|float': function (x, y) {
				return new FloatValue(x.value - y.value);
			},
			'complex': function (x, y) {
				return new ComplexValue(x.value - y.real, -y.imaginary);
			}
		},
		'complex': {
			'int|float': function (x, y) {
				return new ComplexValue(x.real - y.value, x.imaginary);
			},
			'complex': function (x, y) {
				return new ComplexValue(x.real - y.real, x.imaginary - y.imaginary);
			}
		}
	});
}

exports.MinusOperator = MinusOperator;

MinusOperator.prototype = Object.create(BinaryOperatorNode.prototype);


function MultiplyOperator() {
	BinaryOperatorNode.call(this, {
		'bool': {
			'bool': function (x, y) {
				return BooleanValue.unmarshal(x.value && y.value);
			}
		},
		'int': {
			'int': function (x, y) {
				return new IntegerValue(x.value * y.value);
			},
			'float': function (x, y) {
				return new FloatValue(x.value * y.value);
			},
			'complex': function (x, y) {
				return new ComplexValue(x.value * y.real, x.value * y.imaginary);
			}
		},
		'float': {
			'int|float': function (x, y) {
				return new FloatValue(x.value * y.value);
			},
			'complex': function (x, y) {
				return new ComplexValue(x.value * y.real, x.value * y.imaginary);
			}
		},
		'complex': {
			'int|float': function (x, y) {
				return new ComplexValue(x.real * y.value, x.imaginary * y.value);
			},
			'complex': function (x, y) {
				return new ComplexValue(
					x.real * y.real - x.imaginary * y.imaginary,
					x.real * y.imaginary + x.imaginary * y.real
					);
			}
		}
	});
}

exports.MultiplyOperator = MultiplyOperator;

MultiplyOperator.prototype = Object.create(BinaryOperatorNode.prototype);


function DivideOperator() {
	BinaryOperatorNode.call(this, {
		'int': {
			'int': function (x, y) {
				var value = x.value / y.value;
				if (value < 0) {
					return new IntegerValue(Math.ceil(value));
				} else {
					return new IntegerValue(Math.floor(value));
				}
			},
			'float': function (x, y) {
				return new FloatValue(x.value / y.value);
			},
			'complex': function (x, y) {
				return new ComplexValue(
					x.value * y.real / (y.real * y.real + y.imaginary * y.imaginary),
					x.value * y.imaginary / (y.real * y.real + y.imaginary * y.imaginary)
					);
			}
		},
		'float': {
			'int|float': function (x, y) {
				return new FloatValue(x.value / y.value);
			},
			'complex': function (x, y) {
				return new ComplexValue(
					x.value * y.real / (y.real * y.real + y.imaginary * y.imaginary),
					x.value * y.imaginary / (y.real * y.real + y.imaginary * y.imaginary)
					);
			}
		},
		'complex': {
			'int|float': function (x, y) {
				return new ComplexValue(x.real / y.value, x.imaginary / y.value);
			},
			'complex': function (x, y) {
				return new ComplexValue(
					(x.real * y.imaginary + x.imaginary * y.real) / (y.real * y.real + y.imaginary * y.imaginary),
					(x.imaginary * y.real - x.real * y.imaginary) / (y.real * y.real + y.imaginary * y.imaginary)
					);
			}
		}
	});
}

exports.DivideOperator = DivideOperator;

DivideOperator.prototype = Object.create(BinaryOperatorNode.prototype);


function PowerOperator() {
	BinaryOperatorNode.call(this, {
		'int': {
			'int': function (x, y) {
				return new IntegerValue(Math.pow(x.value, y.value));
			},
			'float': function (x, y) {
				return new FloatValue(Math.pow(x.value, y.value));
			}
		},
		'float': {
			'int|float': function (x, y) {
				return new FloatValue(Math.pow(x.value, y.value));
			}
		}
	});
}

exports.PowerOperator = PowerOperator;

PowerOperator.prototype = Object.create(BinaryOperatorNode.prototype);


function ModulusOperator() {
	BinaryOperatorNode.call(this, {
		'int': {
			'int': function (x, y) {
				return new IntegerValue(x.value % y.value);
			},
			'float': function (x, y) {
				return new FloatValue(x.value % y.value);
			}
		},
		'float': {
			'int|float': function (x, y) {
				return new FloatValue(x.value % y.value);
			}
		}
	});
}

exports.ModulusOperator = ModulusOperator;

ModulusOperator.prototype = Object.create(BinaryOperatorNode.prototype);


function LessThanOperator() {
	BinaryOperatorNode.call(this, {
		'bool': {
			'bool': function (x, y) {
				return BooleanValue.unmarshal(!x.value && y.value);
			}
		},
		'int|float': {
			'int|float': function (x, y) {
				return BooleanValue.unmarshal(x.value < y.value);
			}
		}
	});
}

exports.LessThanOperator = LessThanOperator;

LessThanOperator.prototype = Object.create(BinaryOperatorNode.prototype);


function LessThanOrEqualOperator() {
	BinaryOperatorNode.call(this, {
		'bool': {
			'bool': function (x, y) {
				return BooleanValue.unmarshal(!x.value || y.value);
			}
		},
		'int|float': {
			'int|float': function (x, y) {
				return BooleanValue.unmarshal(x.value <= y.value);
			}
		}
	});
}

exports.LessThanOrEqualOperator = LessThanOrEqualOperator;

LessThanOrEqualOperator.prototype = Object.create(BinaryOperatorNode.prototype);


function GreaterThanOperator() {
	BinaryOperatorNode.call(this, {
		'bool': {
			'bool': function (x, y) {
				return BooleanValue.unmarshal(x.value && !y.value);
			}
		},
		'int|float': {
			'int|float': function (x, y) {
				return BooleanValue.unmarshal(x.value > y.value);
			}
		}
	});
}

exports.GreaterThanOperator = GreaterThanOperator;

GreaterThanOperator.prototype = Object.create(BinaryOperatorNode.prototype);


function GreaterThanOrEqualOperator() {
	BinaryOperatorNode.call(this, {
		'bool': {
			'bool': function (x, y) {
				return BooleanValue.unmarshal(x.value || !y.value);
			}
		},
		'int|float': {
			'int|float': function (x, y) {
				return BooleanValue.unmarshal(x.value >= y.value);
			}
		}
	});
}

exports.GreaterThanOrEqualOperator = GreaterThanOrEqualOperator;

GreaterThanOrEqualOperator.prototype = Object.create(BinaryOperatorNode.prototype);


function BitwiseAndOperator() {
	BinaryOperatorNode.call(this, {
		'int': {
			'int': function (x, y) {
				return new IntegerValue(x.value & y.value);
			}
		}
	});
}

exports.BitwiseAndOperator = BitwiseAndOperator;

BitwiseAndOperator.prototype = Object.create(BinaryOperatorNode.prototype);


function BitwiseOrOperator() {
	BinaryOperatorNode.call(this, {
		'int': {
			'int': function (x, y) {
				return new IntegerValue(x.value | y.value);
			}
		}
	});
}

exports.BitwiseOrOperator = BitwiseOrOperator;

BitwiseOrOperator.prototype = Object.create(BinaryOperatorNode.prototype);


function BitwiseXorOperator() {
	BinaryOperatorNode.call(this, {
		'int': {
			'int': function (x, y) {
				return new IntegerValue(x.value ^ y.value);
			}
		}
	});
}

exports.BitwiseXorOperator = BitwiseXorOperator;

BitwiseXorOperator.prototype = Object.create(BinaryOperatorNode.prototype);


function LeftShiftOperator() {
	BinaryOperatorNode.call(this, {
		'int': {
			'int': function (x, y) {
				return new IntegerValue(x.value << y.value);
			}
		}
	});
}

exports.LeftShiftOperator = LeftShiftOperator;

LeftShiftOperator.prototype = Object.create(BinaryOperatorNode.prototype);


function RightShiftOperator() {
	BinaryOperatorNode.call(this, {
		'int': {
			'int': function (x, y) {
				return new IntegerValue(x.value >> y.value);
			}
		}
	});
}

exports.RightShiftOperator = RightShiftOperator;

RightShiftOperator.prototype = Object.create(BinaryOperatorNode.prototype);


function UnsignedRightShiftOperator() {
	BinaryOperatorNode.call(this, {
		'int': {
			'int': function (x, y) {
				return new IntegerValue(x.value >>> y.value);
			}
		}
	});
}

exports.UnsignedRightShiftOperator = UnsignedRightShiftOperator;

UnsignedRightShiftOperator.prototype = Object.create(BinaryOperatorNode.prototype);


function ComparisonOperator() {
	BinaryOperatorNode.call(this, {
		'undefined': {
			'undefined': function () {
				return BooleanValue.TRUE;
			},
			'string': function () {
				return BooleanValue.FALSE;
			},
			'closure': function () {
				return BooleanValue.FALSE;
			},
			'array': function () {
				return BooleanValue.FALSE;
			},
			'object': function () {
				return BooleanValue.FALSE;
			}
		},
		'null': {
			'null': function () {
				return BooleanValue.TRUE;
			},
			'string': function () {
				return BooleanValue.FALSE;
			},
			'closure': function () {
				return BooleanValue.FALSE;
			},
			'array': function () {
				return BooleanValue.FALSE;
			},
			'object': function () {
				return BooleanValue.FALSE;
			}
		},
		'bool': {
			'bool': function (x, y) {
				return BooleanValue.unmarshal(x.value === y.value);
			}
		},
		'int|float': {
			'int|float': function (x, y) {
				return BooleanValue.unmarshal(x.value === y.value);
			},
			'complex': function (x, y) {
				return BooleanValue.unmarshal(x.value === y.real && !y.imaginary);
			}
		},
		'complex': {
			'int|float': function (x, y) {
				return BooleanValue.unmarshal(!x.imaginary && x.real === y.value);
			},
			'complex': function (x, y) {
				return BooleanValue.unmarshal(x.real === y.real && x.imaginary === y.imaginary);
			}
		},
		'string': {
			'undefined|null': function () {
				return BooleanValue.FALSE;
			},
			'string': function (x, y) {
				return BooleanValue.unmarshal(x.value === y.value);
			}
		},
		'array': {
			'undefined|null': function () {
				return BooleanValue.FALSE;
			},
			'array': function (x, y) {
				if (x.array.length !== y.array.length) {
					return BooleanValue.FALSE;
				} else {
					for (var i = 0; i < x.array.length; i++) {
						// TODO
						throw new LambdaInternalError();
					}
					return BooleanValue.TRUE;
				}
			}
		},
		'object': {
			'undefined|null': function () {
				return BooleanValue.FALSE;
			},
			'object': function () {
				// TODO
				throw new LambdaInternalError();
			}
		}
	});
}

exports.ComparisonOperator = ComparisonOperator;

ComparisonOperator.prototype = Object.create(BinaryOperatorNode.prototype);


function NegatedComparisonOperator() {
	BinaryOperatorNode.call(this, {
		'undefined': {
			'undefined': function () {
				return BooleanValue.FALSE;
			},
			'string': function () {
				return BooleanValue.TRUE;
			},
			'closure': function () {
				return BooleanValue.TRUE;
			},
			'array': function () {
				return BooleanValue.TRUE;
			},
			'object': function () {
				return BooleanValue.TRUE;
			}
		},
		'null': {
			'null': function () {
				return BooleanValue.FALSE;
			},
			'string': function () {
				return BooleanValue.TRUE;
			},
			'closure': function () {
				return BooleanValue.TRUE;
			},
			'array': function () {
				return BooleanValue.TRUE;
			},
			'object': function () {
				return BooleanValue.TRUE;
			}
		},
		'bool': {
			'bool': function (x, y) {
				return BooleanValue.unmarshal(x.value !== y.value);
			}
		},
		'int|float': {
			'int|float': function (x, y) {
				return BooleanValue.unmarshal(x.value !== y.value);
			},
			'complex': function (x, y) {
				return BooleanValue.unmarshal(x.value !== y.real || !!y.imaginary);
			}
		},
		'complex': {
			'int|float': function (x, y) {
				return BooleanValue.unmarshal(!!x.imaginary || x.real !== y.value);
			},
			'complex': function (x, y) {
				return BooleanValue.unmarshal(x.real !== y.real || x.imaginary !== y.imaginary);
			}
		},
		'string': {
			'undefined|null': function () {
				return BooleanValue.TRUE;
			},
			'string': function (x, y) {
				return BooleanValue.unmarshal(x.value !== y.value);
			}
		},
		'array': {
			'undefined|null': function () {
				return BooleanValue.TRUE;
			},
			'array': function (x, y) {
				if (x.array.length !== y.array.length) {
					return BooleanValue.TRUE;
				} else {
					for (var i = 0; i < x.array.length; i++) {
						// TODO
						throw new LambdaInternalError();
					}
					return BooleanValue.FALSE;
				}
			}
		},
		'object': {
			'undefined|null': function () {
				return BooleanValue.TRUE;
			},
			'object': function () {
				// TODO
				throw new LambdaInternalError();
			}
		}
	});
}

exports.NegatedComparisonOperator = NegatedComparisonOperator;

NegatedComparisonOperator.prototype = Object.create(BinaryOperatorNode.prototype);


function LogicalAndOperator() {
	BinaryOperatorNode.call(this, {
		'bool': {
			'bool': function (x, y) {
				return BooleanValue.unmarshal(x.value && y.value);
			}
		}
	});
}

exports.LogicalAndOperator = LogicalAndOperator;

LogicalAndOperator.prototype = Object.create(BinaryOperatorNode.prototype);


function LogicalOrOperator() {
	BinaryOperatorNode.call(this, {
		'bool': {
			'bool': function (x, y) {
				return BooleanValue.unmarshal(x.value || y.value);
			}
		}
	});
}

exports.LogicalOrOperator = LogicalOrOperator;

LogicalOrOperator.prototype = Object.create(BinaryOperatorNode.prototype);


function LogicalXorOperator() {
	BinaryOperatorNode.call(this, {
		'bool': {
			'bool': function (x, y) {
				return BooleanValue.unmarshal(x.value !== y.value);
			}
		}
	});
}

exports.LogicalXorOperator = LogicalXorOperator;

LogicalXorOperator.prototype = Object.create(BinaryOperatorNode.prototype);

function DefaultContext() {
	function evaluate(Operator) {
		return (new Operator()).evaluate(Context.EMPTY);
	}

	Context.call(this, {
		'typeof': evaluate(TypeOfOperator),
		'not': evaluate(LogicalNotOperator),
		'~': evaluate(BitwiseNotOperator),
		'+': evaluate(PlusOperator),
		'-': evaluate(MinusOperator),
		'*': evaluate(MultiplyOperator),
		'/': evaluate(DivideOperator),
		'**': evaluate(PowerOperator),
		'%': evaluate(ModulusOperator),
		'<': evaluate(LessThanOperator),
		'<=': evaluate(LessThanOrEqualOperator),
		'>': evaluate(GreaterThanOperator),
		'>=': evaluate(GreaterThanOrEqualOperator),
		'&': evaluate(BitwiseAndOperator),
		'|': evaluate(BitwiseOrOperator),
		'^': evaluate(BitwiseXorOperator),
		'<<': evaluate(LeftShiftOperator),
		'>>': evaluate(RightShiftOperator),
		'>>>': evaluate(UnsignedRightShiftOperator),
		'=': evaluate(ComparisonOperator),
		'!=': evaluate(NegatedComparisonOperator),
		'and': evaluate(LogicalAndOperator),
		'or': evaluate(LogicalOrOperator),
		'xor': evaluate(LogicalXorOperator)
	});
}

exports.DefaultContext = DefaultContext;

DefaultContext.prototype = Object.create(Context.prototype);

function Lexer(input) {
	var token, label;

	function match(re) {
		var result = re.exec(input);
		if (result) {
			label = result[0];
			input = input.substr(label.length);
			return true;
		} else {
			return false;
		}
	}

	var keywords = [
		'null',
		'undefined',
		'true',
		'false',
		'not',
		'and',
		'or',
		'xor',
		'fix',
		'this',
		'new',
		'bool',
		'unknown',
		'int',
		'float',
		'complex',
		'string',
		'regex',
		'let',
		'in',
		'if',
		'then',
		'else',
		'throw',
		'try',
		'catch',
		'finally',
		'error',
		'throws'
	];

	function next() {
		if (match(/^(\s|(#.*\n))+/)) {
			return next();
		} else if (match(/^[A-Za-z_\$][A-Za-z0-9_\$]*/)) {
			if (keywords.indexOf(label) < 0) {
				return token = 'identifier';
			} else {
				return token = 'keyword:' + label;
			}
		} else if (match(/^\-\>/)) {
			return token = 'arrow';
		} else if (match(/^\=\>/)) {
			return token = 'fat-arrow';
		} else if (match(/^\,/)) {
			return token = 'comma';
		} else if (match(/^\./)) {
			return token = 'point';
		} else if (match(/^\:/)) {
			return token = 'colon';
		} else if (match(/^\(/)) {
			return token = 'left';
		} else if (match(/^\)/)) {
			return token = 'right';
		} else if (match(/^\[/)) {
			return token = 'left-square';
		} else if (match(/^\]/)) {
			return token = 'right-square';
		} else if (match(/^\{/)) {
			return token = 'left-curly';
		} else if (match(/^\}/)) {
			return token = 'right-curly';
		} else if (match(/^(0|[1-9][0-9]*)(\.[0-9]+)?i\b/)) {
			label = parseFloat(label);
			return token = 'complex';
		} else if (match(/^(0|[1-9][0-9]*)\.[0-9]+\b/)) {
			label = parseFloat(label);
			return token = 'float';
		} else if (match(/^0[xX][0-9a-fA-F]+\b/)) {
			label = parseInt(label, 16);
			return token = 'integer';
		} else if (match(/^0[0-7]*\b/)) {
			label = parseInt(label, 8);
			return token = 'integer';
		} else if (match(/^(0|[1-9][0-9]*)\b/)) {
			label = parseInt(label, 10);
			return token = 'integer';
		} else if (match(/^\"([^"\\]|\\\\|\\["bfnrtv])*\"/)) {
			label = label
				.replace(/^\"|\"$/g, '')
				.replace(/\\\\/g, '\\')
				.replace(/\\\"/g, '\"')
				.replace(/\\b/g, '\b')
				.replace(/\\f/g, '\f')
				.replace(/\\n/g, '\n')
				.replace(/\\r/g, '\r')
				.replace(/\\t/g, '\t')
				.replace(/\\v/g, '\v');
			return token = 'string';
		} else if (match(/^>>>/)) {
			return token = 'symbol';
		} else if (match(/^(<<|>>|\!\=|<\=|>\=|\*\*)/)) {
			return token = 'symbol';
		} else if (match(/^(<|>|\+|\-|\/|\%|\~|\&|\||\^)/)) {
			return token = 'symbol';
		} else if (match(/^\*/)) {
			return token = 'asterisk';
		} else if (match(/^\=/)) {
			return token = 'equal';
		} else if (/^$/.test(input)) {
			label = null;
			return token = 'end';
		} else {
			throw new LambdaSyntaxError();
		}
	}

	next();

	this.next = next;

	this.end = function () {
		return token === 'end';
	};

	this.token = function () {
		return token;
	};

	this.label = function () {
		return label;
	};

	this.expect = function (expectedToken) {
		if (expectedToken !== token) {
			throw new LambdaSyntaxError();
		}
		var currentLabel = label;
		next();
		return currentLabel;
	};
}

exports.Lexer = Lexer;

function Parser(input) {
	this.lexer = new Lexer(input);
}

exports.Parser = Parser;

Parser.prototype.parseInteger = function () {
	return new LiteralNode(new IntegerValue(this.lexer.expect('integer')));
};

Parser.prototype.parseComplex = function () {
	return new LiteralNode(new ComplexValue(0, this.lexer.expect('complex')));
};

Parser.prototype.parseFloat = function () {
	return new LiteralNode(new FloatValue(this.lexer.expect('float')));
};

Parser.prototype.parseString = function () {
	return new LiteralNode(new StringValue(this.lexer.expect('string')));
};

Parser.prototype.parseVariable = function () {
	var label = this.lexer.label();
	this.lexer.next();
	return new VariableNode(label);
};

Parser.prototype.parseClass0 = function () {
	switch (this.lexer.token()) {
	case 'keyword:null':
		this.lexer.next();
		return new LiteralNode(NullValue.INSTANCE);
	case 'keyword:undefined':
		this.lexer.next();
		return new LiteralNode(UndefinedValue.INSTANCE);
	case 'keyword:true':
		this.lexer.next();
		return new LiteralNode(BooleanValue.TRUE);
	case 'keyword:false':
		this.lexer.next();
		return new LiteralNode(BooleanValue.FALSE);
	case 'integer':
		return this.parseInteger();
	case 'complex':
		return this.parseComplex();
	case 'float':
		return this.parseFloat();
	case 'string':
		return this.parseString();
	case 'identifier':
	case 'keyword:typeof':
	case 'keyword:not':
	case 'keyword:and':
	case 'keyword:or':
	case 'keyword:xor':
	case 'symbol':
	case 'equal':
	case 'asterisk':
		return this.parseVariable();
	case 'keyword:fix':
		this.lexer.next();
		return FixNode.INSTANCE;
	case 'keyword:this':
		this.lexer.next();
		return ThisNode.INSTANCE;
	case 'keyword:error':
		this.lexer.next();
		return ErrorNode.INSTANCE;
	case 'left':
		this.lexer.next();
		var node = this.parseClass3(['right']);
		this.lexer.next();
		return node;
	case 'left-curly':
		var expressions = [];
		if (this.lexer.next() !== 'right-curly') {
			expressions.push(this.parseClass3(['comma', 'right-curly']));
			while (this.lexer.token() === 'comma') {
				this.lexer.next();
				expressions.push(this.parseClass3(['comma', 'right-curly']));
			}
		}
		this.lexer.next();
		return new ArrayLiteralNode(expressions);
	default:
		throw new LambdaSyntaxError();
	}
};

Parser.prototype.parseSubscriptOrFieldAccess = function (node) {
	while (true) {
		switch (this.lexer.token()) {
		case 'point':
			var token = this.lexer.next();
			if (token !== 'identifier' && !token.match(/^keyword\:/)) {
				throw new LambdaSyntaxError();
			}
			node = new FieldAccessNode(node, this.lexer.expect(token));
			break;
		case 'left-square':
			this.lexer.next();
			var index = this.parseClass3(['right-square']);
			this.lexer.next();
			node = new SubscriptNode(node, index);
			break;
		default:
			return node;
		}
	}
};

Parser.prototype.parseClass1 = function () {
	return this.parseSubscriptOrFieldAccess(this.parseClass0());
};

Parser.prototype.parseLambdaPartial = function (terminators) {
	var name = this.lexer.expect('identifier');
	switch (this.lexer.token()) {
	case 'comma':
		this.lexer.next();
		return new LambdaNode(name, this.parseLambdaPartial(terminators));
	case 'arrow':
		this.lexer.next();
		return new LambdaNode(name, this.parseClass3(terminators));
	default:
		throw new LambdaSyntaxError();
	}
};

Parser.prototype.parseLambdaOrVariable = function (terminators) {
	var name = this.lexer.label();
	switch (this.lexer.next()) {
	case 'comma':
		if (terminators.contains('comma')) {
			return this.parseSubscriptOrFieldAccess(new VariableNode(name));
		} else {
			this.lexer.next();
			return new LambdaNode(name, this.parseLambdaPartial(terminators));
		}
		break;
	case 'arrow':
		this.lexer.next();
		return new LambdaNode(name, this.parseClass3(terminators));
	default:
		return this.parseSubscriptOrFieldAccess(new VariableNode(name));
	}
};

Parser.prototype.parseLetPartial = function (terminators) {
	var names = [this.lexer.expect('identifier')];
	while (this.lexer.token() === 'point') {
		var token = this.lexer.next();
		if (token !== 'identifier' && !token.match(/^keyword\:/)) {
			throw new LambdaSyntaxError();
		} else {
			names.push(this.lexer.expect(token));
		}
	}
	this.lexer.expect('equal');
	var expression = this.parseClass3(['comma', 'keyword:in']);
	switch (this.lexer.token()) {
	case 'comma':
		this.lexer.next();
		return new LetNode(names, expression, this.parseLetPartial(terminators));
	case 'keyword:in':
		this.lexer.next();
		return new LetNode(names, expression, this.parseClass3(terminators));
	default:
		throw new LambdaSyntaxError();
	}
};

Parser.prototype.parseLet = function (terminators) {
	this.lexer.expect('keyword:let');
	return this.parseLetPartial(terminators);
};

Parser.prototype.parseIf = function (terminators) {
	this.lexer.expect('keyword:if');
	var condition = this.parseClass3(['keyword:then']);
	this.lexer.next();
	var thenExpression = this.parseClass3(['keyword:else']);
	this.lexer.next();
	return new IfNode(condition, thenExpression, this.parseClass3(terminators));
};

Parser.prototype.parseThrow = function (terminators) {
	this.lexer.expect('keyword:throw');
	return new ThrowNode(this.parseClass3(terminators));
};

Parser.prototype.parseTry = function (terminators) {
	this.lexer.expect('keyword:try');
	var tryExpression = this.parseClass3(['keyword:catch', 'keyword:finally']);
	switch (this.lexer.token()) {
	case 'keyword:catch':
		this.lexer.next();
		var catchExpression = this.parseClass3(terminators.union(['keyword:finally']));
		if (this.lexer.token() === 'keyword:finally') {
			this.lexer.next();
			return new TryCatchFinallyNode(tryExpression, catchExpression, this.parseClass3(terminators));
		} else if (terminators.contains(this.lexer.token())) {
			return new TryCatchNode(tryExpression, catchExpression);
		}
		throw new LambdaSyntaxError();
	case 'keyword:finally':
		this.lexer.next();
		return new TryFinallyNode(tryExpression, this.parseClass3(terminators));
	default:
		throw new LambdaSyntaxError();
	}
};

Parser.prototype.parseClass2 = function (terminators) {
	switch (this.lexer.token()) {
	case 'identifier':
		return this.parseLambdaOrVariable(terminators);
	case 'keyword:let':
		return this.parseLet(terminators);
	case 'keyword:if':
		return this.parseIf(terminators);
	case 'keyword:throw':
		return this.parseThrow(terminators);
	case 'keyword:try':
		return this.parseTry(terminators);
	default:
		return this.parseClass1(terminators);
	}
};

Parser.prototype.parseClass3 = function (terminators) {
	var node = this.parseClass2(terminators);
	while (!terminators.contains(this.lexer.token())) {
		node = new ApplicationNode(node, this.parseClass2(terminators));
	}
	return node;
};

Parser.prototype.parse = function () {
	return this.parseClass3(['end']);
};

Closure.prototype.prototype = new Context();

Closure.prototype.prototype = Closure.prototype.prototype.addAll({
	length: LazyValue.unmarshal(function () {
		return this.length;
	}),
	apply: Closure.unmarshal(function (parameters) {
		return this.apply(null, parameters);
	})
});

ThisNode.INSTANCE = new ThisNode();

ErrorNode.INSTANCE = new ErrorNode();

FixNode.Z_COMBINATOR = (new LambdaNode('f', new ApplicationNode(
	new LambdaNode('x', new ApplicationNode(
		new VariableNode('f'),
		new LambdaNode('v', new ApplicationNode(
			new ApplicationNode(
				new VariableNode('x'),
				new VariableNode('x')
				),
			new VariableNode('v')
			))
		)),
	new LambdaNode('x', new ApplicationNode(
		new VariableNode('f'),
		new LambdaNode('v', new ApplicationNode(
			new ApplicationNode(
				new VariableNode('x'),
				new VariableNode('x')
				),
			new VariableNode('v')
			))
		))
	))).evaluate(Context.EMPTY);

FixNode.INSTANCE = new FixNode();

DefaultContext.INSTANCE = new DefaultContext();

ArrayValue.prototype.prototype = NativeArrayValue.prototype.prototype = new Context({
	length: LazyValue.unmarshal(function () {
		return this.length;
	}),
	slice: Closure.unmarshal(function (begin, end) {
		return this.slice(begin, end);
	}),
	concat: Closure.unmarshal(function (other) {
		return this.concat(other);
	}),
	indexOf: Closure.unmarshal(function (value) {
		return this.indexOf(value);
	}),
	lastIndexOf: Closure.unmarshal(function (value) {
		return this.lastIndexOf(value);
	}),
	join: Closure.unmarshal(function (glue) {
		return this.join(glue);
	}),
	forEach: Closure.unmarshal(function (callback) {
		this.forEach(function (element) {
			callback(element);
		});
	}),
	some: Closure.unmarshal(function (callback) {
		return this.some(function (element) {
			return callback(element);
		});
	}),
	every: Closure.unmarshal(function (callback) {
		return this.every(function (element) {
			return callback(element);
		});
	}),
	filter: Closure.unmarshal(function (callback) {
		return this.filter(function (element) {
			return callback(element);
		});
	}),
	map: Closure.unmarshal(function (callback) {
		return this.map(function (element) {
			return callback(element);
		});
	}),
	reduce: Closure.unmarshal(function (callback, value) {
		return this.reduce(function (value, element) {
			return callback(value, element);
		}, value);
	}),
	reduceRight: Closure.unmarshal(function (callback, value) {
		return this.reduceRight(function (value, element) {
			return callback(value, element);
		}, value);
	}),
	reverse: LazyValue.unmarshal(function () {
		return this.slice().reverse();
	}),
	sort: Closure.unmarshal(function (compare) {
		return this.sort(function (a, b) {
			if (!compare(a, b)) {
				return 1;
			} else if (!compare(b, a)) {
				return -1;
			} else {
				return 0;
			}
		});
	})
});
